from os.path import join
from subprocess import call, STDOUT, DEVNULL


def command_generator(arguments: dict) -> str:
    pgi, tests, args, args_unknown = arguments

    # get arguments and construct the behave command
    results_json_file = join(args.output, f'{pgi}.json')
    user_defs = ' '.join('-D "{}"'.format(arg) for arg in args.D) if args.D else ''
    tags = ' '.join('--tags "{}"'.format(','.join(t for t in g)) for g in args.tags)
    stage = f'--stage="{args.stage}"' if args.stage else ''
    output = f'-f json -o "{join(args.output, str(pgi))}.json"'

    command = (
        f'behave {stage} -D "parallel={pgi}" {user_defs}'
        f' {tags} {output} {" ".join(args_unknown)} '
    )

    feature_files = set(thing.path.split(':')[0] for thing in tests)
    if args.unit not in ['feature', 'example', 'scenario']:
        args.unit = 'scenario'

    if args.unit == 'feature':
        command += ' '.join(feature_files)
    else:
        command += ' '.join(thing.path for thing in tests)

    if args.dry_run:
        print(command)
        return ''

    if args.verbose:
        print(command)
        call(command, shell=True)
    else:
        call(command, shell=True, stdout=DEVNULL, stderr=STDOUT)

    return results_json_file



from random import shuffle
from enum import StrEnum


class Unit(StrEnum):
    SCENARIO = 'scenario'
    FEATURE = 'feature'
    EXAMPLE = 'example'


def matches_tag_expression(scenario_tags, tag_expression):
    # handle trivial cases
    if not tag_expression:
        return True
    if not scenario_tags:
        return False

    # separate included and excluded tags
    excluded_tags = [t for e in tag_expression for t in e if t.startswith('~')]
    tag_expression = [[t for t in group if t not in excluded_tags] for group in tag_expression]

    # boolean logic to figure out if the test should run
    matches = []
    for expr in tag_expression:
        res = all([t in scenario_tags for t in expr])
        matches.append(res)
    return any(matches) and all([t.replace('~', '') not in scenario_tags for t in excluded_tags if t.startswith('~')])


def get_grouped_tests(features, tags, num_split, unit=None, shuffle_paths=True):
    # split as evenly as possible
    # it's ugly, sorry
    paths = []
    all_tests = []

    for feature in features:

        if unit == Unit.SCENARIO:
            all_tests.extend(feature.get_testable_tests(tags))
        elif unit == Unit.FEATURE:
            all_tests.append(feature)
        elif unit == Unit.EXAMPLE:
            tests = feature.get_testable_tests(tags)
            scenarios = [t for t in tests if t.keyword.lower() == 'scenario']
            examples = [row for test in tests for table in test.tables for row in table.rows[1:]
                        if test.keyword.lower() == 'scenario outline']
            all_tests.extend(scenarios)
            all_tests.extend(examples)

    inc_other = -(-len(all_tests) // num_split)

    if shuffle_paths:
        shuffle(paths)
        shuffle(all_tests)
    test_groups = [all_tests[i:i + inc_other] for i in range(0, len(all_tests), inc_other)]

    return [[i, group] for i, group in enumerate(test_groups)]



import re
from anubis.gherkin import Background, Feature, Row, Step, Test, Table
from glob import glob
from os.path import join
from typing import Tuple, Union


breaking_kwds = [
    'Feature',
    'Rule',
    # 'Example',
]

section_kwds = [
    'Background',
    'Scenario Outline',
    'Scenario Template',
    'Scenario',
]

step_kwds = [
    'Given',
    'When',
    'Then',
    'And',
    'But',
    r'\*',
]

primary_kw = [
    'Scenarios',
    'Examples',
]

meta_kw = ['|', '@']
ignore_kw = ['"""', '#']


def __get_kw_and_description(line: str) -> Tuple[Union[str, None], Union[str, None]]:
    kwds = breaking_kwds + section_kwds + step_kwds + primary_kw
    pattern = re.compile(r'^(?P<kw>' + "|".join(kw for kw in kwds) + ')(:)?(?P<description>.*)$')
    match = re.match(pattern, line)
    if match:
        return match.group('kw'), match.group('description').lstrip().rstrip()
    return None, None


def better_gherkin_parser(fp):
    with open(fp, 'r') as f:
        lines = [line.lstrip().replace('\n', '') for line in f.readlines()]
    lines.reverse()

    feature = None
    tests = []
    steps = []
    background = Background()
    current_test = Test()
    current_test_tables = []
    current_table = Table()

    for ln, line in enumerate(lines):
        line_num = str(len(lines) - ln)

        if any([line.startswith(kw) for kw in step_kwds]):
            kw, desc = __get_kw_and_description(line)
            steps.append(Step(desc, fp, kw, line_num, current_table))
            current_table = Table()
        elif any([line.startswith(kw) for kw in section_kwds]) or line.startswith('Background'):
            kw, desc = __get_kw_and_description(line)
            steps.reverse()

            current_test.file = fp
            current_test.line = line_num
            current_test.steps = steps
            current_test.path = f'{fp}:{line_num}'

            if kw != 'Background':
                current_test.description = desc
                current_test.keyword = kw
                current_test.tables = current_test_tables
                tests.append(current_test)
                current_test = Test()
            else:
                background.steps = steps
                background.description = desc

            current_test_tables = []
            current_table = Table()
            steps = []
        elif line.startswith('@'):
            (tests[-1] if ln != len(lines) - 1 else feature).tags.extend([tag.replace('@', '') for tag in line.split()])
        elif line.startswith('|'):
            current_table.rows.append(
                Row([data.strip() for data in line.split('|') if data], fp, line_num))
        elif line.startswith('Examples'):
            current_table.rows.reverse()
            current_test_tables.append(current_table)
            current_table = Table()
        elif line.startswith('Feature'):
            tests.reverse()
            feature = Feature(
                description=__get_kw_and_description(line)[1],
                background=background,
                file=fp,
                line=line_num,
                tests=tests
            )
    return feature


def get_parsed_gherkin(feature_dir):
    all_paths = []
    parsed_gherkin = []

    for directory in feature_dir:
        all_paths.extend(glob(join(directory, '**', '*.feature'), recursive=True))

    for fp in all_paths:
        parsed_gherkin.append(better_gherkin_parser(fp))
    return parsed_gherkin



import json
import os
import logging
from json.decoder import JSONDecodeError
from pathlib import Path
from time import gmtime, strftime
import logging


GREEN_START = '\033[92m'
GREY_START = '\033[90m'
RED_START = '\033[91m'
END_COLOR = '\033[0m'
spc = '  '
linesep = os.linesep
tab = '\t'


root_logger = logging.getLogger()


def __write_scenario_junit(scenario):
    duration = 0
    formatted_steps = ''
    try:
        for step in scenario["steps"]:
            line = os.linesep + tab + step['keyword'] + ' '  + step['name'].replace('"', '&#34;').replace('&', '&#38;')
            if 'result' in step:
                duration += step['result']['duration']
                line += " ... " + step["result"]["status"] + " in " + str("{:.3f}s".format(step["result"]["duration"]))
            formatted_steps += line
        start = f'\n<testcase classname="{scenario["location"]}" name="{scenario["name"]}" status="{scenario["status"]}" time="{duration}">'

        if scenario['status'] == 'passed':
            failure_message = ''
        else:
            failed_step = next((s for s in scenario['steps'] if 'result' in s and s['result']['status'] == 'failed'), scenario['steps'][-1])

            try:
                error_message = failed_step["result"]["error_message"]
                duration = failed_step['result']['duration']
            except KeyError as e:
                logging.exception(f'Error writing scenario steps to junit:\n\t{scenario}')
                error_message = '<could not find error message>'
                duration = 0.0

            failure_message = (
                '\n<failure>\n'
                '<![CDATA[\n'
                f'Failing step: {failed_step["keyword"]} {failed_step["name"]} ... failed in {"{:.3f}s".format(duration)}\n'
                f'Location: {failed_step["location"]}\n'
                f'{linesep.join(error_message) if type(error_message) is list else error_message}\n'
                ']]>\n'
                '</failure>\n'
            )

        output = (
            '\n<system-out>\n'
            '<![CDATA[]\n'
            f'@scenario.begin\n\n'
            f'{"".join(["@" + t + " " for t in scenario["tags"]])}\n'
            f'Scenario: {scenario["name"]}'
            f'{formatted_steps}\n'
            '@scenario.end\n'
            '--------------------------------------------------------------------------------\n'
            ']]>\n'
            '</system-out>\n'
            '</testcase>\n')

        return start + failure_message + output
    except Exception as e:
        logging.error(f'Error writing scenario to junit:\n\t{e}')


def __get_widths(s, p, f, hide_sum, hide_passed, hide_failed, extra_width=0):
    """This does all the terrible work of finding the widths of lines for output"""
    width_p_name = max([len(line['name']) for line in p]) if p else 0
    width_p_location = max([len(line['location']) for line in p]) if p else 0
    width_f_name = max([len(line['name']) for line in f]) if f else 0
    width_f_location = max([len(line['location']) for line in f]) if f else 0
    width_s = max([len(line) for line in s])

    return (
        max(
            (width_p_name + width_p_location) * int(not hide_passed),
            (width_f_name + width_f_location) * int(not hide_failed),
            width_s * int(not hide_sum)) + extra_width,
        width_p_name,
        width_p_location,
        width_f_name,
        width_f_location,
        width_s
    )


def handle_passing_failing_scenarios(is_passing, width, scenarios):
    dc = RED_START
    cc = GREY_START
    end = END_COLOR
    if is_passing:
        dc = GREEN_START

    details = [
        {
            'scenario': f'{spc}{dc}●{end}{spc}{s["name"].ljust(width)}{spc}{cc}',
            'location': f'# {s["location"]}{end}'
        }
        for s in scenarios
    ]
    details.sort(key=lambda x: x['location'])

    return '\n'.join([s['scenario'] + s['location'] for s in details])


def write_result_aggregate(files: list, aggregate_out_file):
    agg_fp = Path(aggregate_out_file)
    aggregate = []

    for fp in files:
        try:
            with open(fp, 'r') as f:
                current_file_data = json.load(f)
        except (FileNotFoundError, JSONDecodeError):
            current_file_data = []
        aggregate += current_file_data

    with agg_fp.open('w+', encoding='utf-8') as f:
        f.write(json.dumps(aggregate))


def get_result_values(aggregate_file):
    try:
        with open(aggregate_file) as f:
            res = json.load(f)
    except FileNotFoundError:
        return 0, 0, 0

    statuses = []
    for feature in res:
        if 'elements' in feature:
            for scenario in feature['elements']:
                if scenario['type'] != 'background':
                    statuses.append(scenario['status'])

    passed = statuses.count('passed')
    failed = statuses.count('failed')
    total = passed + failed
    return passed, failed, total


def print_result_summary(args_known, user_defs, start, end, num_passed, num_failed):
    # set up a bunch of variables
    total = num_passed + num_failed
    pass_rate = f'{num_passed / total * 100:.2f}%' if total > 0 else 'n/a'
    fail_rate = f'{num_failed / total * 100:.2f}%' if total > 0 else 'n/a'
    env_txt = f'{spc}Env:     <{args_known.env}>'
    tag_txt = f'{spc}Tags:    <{[t for t in args_known.tags] if args_known.tags else "n/a"}>'
    res_txt = f'{spc}Output:  <{args_known.output}>'
    pass_txt = f'{spc}Passed:  <{num_passed}, {pass_rate}>'
    fail_txt = f'{spc}Failed:  <{num_failed}, {fail_rate}>'
    run_txt = f'{spc}Runtime: <{strftime("%Hh %Mm %Ss", gmtime((end - start).total_seconds()))}>'

    with open(args_known.aggregate) as f:
        res = json.load(f)

    summary = [pass_rate, fail_rate, env_txt, tag_txt, res_txt, pass_txt, fail_txt, run_txt]
    passed = []
    failed = []
    for feature in res:
        if 'elements' in feature:
            for element in feature['elements']:
                if element['type'] != 'background' and element['status'].lower() == 'passed':
                    passed.append(element)
                if element['type'] != 'background' and element['status'].lower() == 'failed':
                    failed.append(element)

    width, width_pn, width_pl, width_fn, width_fl, width_sum = __get_widths(
        summary, passed, failed, args_known.hide_passed, args_known.hide_failed, args_known.hide_summary, 4 * len(spc))

    if not args_known.hide_passed and num_passed:
        passing_details = handle_passing_failing_scenarios(True, width_pn, passed)
        print('--- Passing scenarios '.ljust(width, '-') + '\n' + passing_details + '\n')
    if not args_known.hide_failed and num_failed:
        failing_details = handle_passing_failing_scenarios(False, width_fn, failed)
        print('--- Failing scenarios '.ljust(width, '-') + '\n' + failing_details + '\n')
    if not args_known.hide_summary:
        print('--- ♡𓃥♡ SUMMARY ♡𓃥♡ '.ljust(width, '-'))
        print(env_txt)
        print(res_txt)
        print(tag_txt)
        print(pass_txt)
        print(fail_txt)
        print(run_txt, '\n')


def write_junit(json_aggregate, output_file):
    with open(json_aggregate) as f:
        res = json.load(f)

    statuses = []
    for feature in res:
        if 'elements' in feature:
            for scenario in feature['elements']:
                if scenario['type'] != 'background':
                    statuses.append(scenario['status'])

    passed = statuses.count('passed')
    failed = statuses.count('failed')
    total = passed + failed

    start = ('<?xml version="1.0" encoding="UTF-8" ?>\n'
             f'<testsuite name="Test Run" tests="{total}" failures="{failed}">')

    scenarios = []
    for feature in res:
        if 'elements' in feature:
            for scenario in feature['elements']:
                bg_steps = [] if scenario['type'].lower() != 'background' else scenario['steps']

                if scenario['type'].lower() != 'background':
                    bg_steps += scenario['steps']
                    scenario['steps'] = bg_steps
                    scenarios.append(scenario)

    try:  # todo - make all of this more stable; logging
        junit_res = [__write_scenario_junit(scen) for scen in scenarios]
        end = '\n</testsuite>'
        final = start + "".join(junit_res) + end

        with open(output_file, 'w+') as f:
            f.writelines(final)
    except Exception as e:
        logging.exception(f'Error writing junit file \n{e}')
        print('!!something went wrong writing to the junit out file!!')



import argparse
import os


def parse_arguments():
    parser = argparse.ArgumentParser('Running in parallel mode')
    # anubis-specific stuff
    parser.add_argument('--aggregate', required=False, default='aggregate.json')
    parser.add_argument('--env', required=True, type=str)
    parser.add_argument('--feature_dir', required=False, default=['features'], nargs='*')
    parser.add_argument('--junit', required=False, type=str, default='results.xml')
    parser.add_argument('--output', required=False, default='.tempoutput')
    parser.add_argument('--pass-threshold', required=False, type=float, default=1.0)
    parser.add_argument('--processes', required=False, default=5, type=int)
    parser.add_argument('--unit', required=False, type=str, default='scenario')
    parser.add_argument('--log-file', type=str, default='latest.log')

    # flags
    parser.add_argument('--dry-run', required=False, action='store_true')
    parser.add_argument('--hide-failed', required=False, action='store_true')
    parser.add_argument('--hide-passed', required=False, action='store_true')
    parser.add_argument('--hide-summary', required=False, action='store_true')
    parser.add_argument('--pass-with-no-tests', required=False, action='store_true')
    parser.add_argument('--verbose', required=False, action='store_true')

    # sent directly to behave
    parser.add_argument('--stage', required=True, type=str)
    parser.add_argument('--tags', required=False, nargs='*', action='append', default=[])
    parser.add_argument('-D', required=False, nargs='*')

    known, unknown = parser.parse_known_args()

    # format anything that needs to be formatted
    # output files
    known.output = os.path.join(os.getcwd(), known.output)
    known.aggregate = os.path.join(known.output, known.aggregate)
    known.junit = os.path.join(known.output, known.junit)
    known.log_file = os.path.join(known.output, known.log_file)

    # misc updates to D
    known.D.append(f'env="{known.env}"')

    return known, unknown



from uuid import uuid4


class Row:
    def __init__(
            self,
            cells:    list = None,
            file:     str = None,
            line:     str = None
    ):
        self.cells    = [] if cells is None else cells
        self.file     = file if file is not None else ''
        self.line     = line if line is not None else ''
        self.path     = f'{self.file}:{self.line}'


class Table:
    def __init__(
            self,
            description: str = None,
            file:        str = None,
            line:        str = None,
            rows:        list[Row] = None
    ):
        self.description = '' if description is None else description
        self.file        = '' if file is None else file
        self.line        = '' if line is None else line
        self.id          = uuid4()
        self.path        = f'{self.file}:{self.line}' if self.file and self.line else ''
        self.rows        = [] if rows is None else rows

    def __bool__(self):
        return bool(self.rows)


class SimpleGherkin:
    def __init__(
            self,
            description: str = None,
            file:        str = None,
            keyword:     str = None,
            line:        str = None
    ):
        self.description = '' if description is None else description
        self.id          = uuid4()
        self.file        = '' if file is None else file
        self.keyword     = '' if keyword is None else keyword
        self.line        = '' if line is None else line
        self.path        = f'{self.file}:{self.line}' if self.file and self.line else ''

    def __repr__(self):
        return f'{self.keyword}: {self.description}'

    def __bool__(self):
        return bool(self.keyword and self.description)


class Step(SimpleGherkin):
    def __init__(
            self,
            description: str   = None,
            file:        str   = None,
            keyword:     str   = None,
            line:        str   = None,
            table:       Table = None
    ):
        super().__init__(description, file, keyword, line)
        self.table = Table() if table is None else table


class Background(SimpleGherkin):
    def __init__(
            self,
            description: str = None,
            file:        str = None,
            line:        str = None,
            steps:       list[Step] = None
    ):
        super().__init__(description, file, 'Background', line)
        self.steps = [] if steps is None else steps


class Test(SimpleGherkin):
    def __init__(
            self,
            background:  Background = None,
            description: str = None,
            file:        str = None,
            keyword:     str = None,
            line:        str = None,
            tags:        list = None,
            steps:       list[Step] = None,
            tables:      list[Table] = None
    ):
        super().__init__(description, file, keyword, line)
        self.background = Background() if background is None else background
        self.tags       = [] if tags is None else tags
        self.tables     = [] if tables is None else tables
        self.steps      = self.background.steps + ([] if steps is None else steps)
        self.path       = f'{self.file}:{self.line}' if self.file and self.line else ''

    def is_testable_based_on_tags(self, tag_expression):
        # compare tags to tag expression
        if not tag_expression:
            return True
        if not self.tags:
            return False

        # separate included and excluded tags
        excluded_tags = [t for e in tag_expression for t in e if t.startswith('~')]
        tag_expression = [[t for t in group if t not in excluded_tags] for group in tag_expression]

        # boolean logic to figure out if the test should run
        matches = []
        for expr in tag_expression:
            res = all([t in self.tags for t in expr])
            matches.append(res)
        return any(matches) and all([t.replace('~', '') not in self.tags for t in excluded_tags if t.startswith('~')])

    def get_test_lines(self):
        entire_test = [self.location]
        examples = [row.location for table in self.tables for row in table.rows[1:]]  # exclude the header row

        return {
            'test': entire_test,
            'examples': examples
        }


class Feature:
    def __init__(
            self,
            description: str = None,
            background:  Background = None,
            file:        str = None,
            line:        str = None,
            tags:        list = None,
            tests:       list[Test] = None
    ):
        self.background  = Background() if background is None else background
        self.description = '' if description is None else description
        self.id          = uuid4()
        self.keyword     = 'Feature'
        self.file        = file
        self.line        = line
        self.path        = f'{self.file}:{self.line}' if self.file and self.line else ''
        self.tags        = [] if tags is None else tags
        self.tests       = [] if tests is None else tests

        for test in self.tests:
            test.background = self.background
            test.tags.extend(self.tags)

    def __repr__(self):
        return f'Feature: {self.description}'

    def __str__(self):
        return f'Feature: {self.description}\n' + '\n'.join([f'\t{t.keyword}: {t.description}' for t in self.tests])

    def get_testable_tests(self, tag_expression):
        return [t for t in self.tests if t.is_testable_based_on_tags(tag_expression)]



# __main__.py
import os
import sys
import shutil
from datetime import datetime
from multiprocessing import Pool
from random import choice
from anubis import feature_file_parser, arg_parser_main, feature_splitter, results
from anubis.parallelizer import command_generator
from anubis.copy import art, power
import logging


def main():
    # Misc Setup ---------------------------------------------------------------------------------
    start = datetime.now()
    args, args_unknown = arg_parser_main.parse_arguments()
    print(
        choice(art) + '\nRunning tests with Anubis  |  powered by ' + choice(power),
        end='\n\n') if args.verbose else None

    # Set up output dirs and files ---------------------------------------------------------------------------------
    # create a directory that will contain results and be exported
    print(f'--- Setting Up Output\n\tSending output to <{args.output}>') if args.verbose else None

    if os.path.isdir(args.output):
        shutil.rmtree(args.output)
    os.makedirs(args.output, exist_ok=True)

    # set up logging
    logging.basicConfig(
        filename=args.log_file,
        filemode='w',
        level=logging.DEBUG,
        format='[%(asctime)s] [%(levelname)s] %(message)s',
        datefmt='%H:%M:%S'
    )
    root_logger = logging.getLogger()
    root_logger.info('Args: \n\t' + "\n\t".join(a.__repr__() for a in args._get_kwargs()))

    # Group features and tests ---------------------------------------------------------------------------------
    if args.verbose:
        print('\n--- Grouping Tests')
        print('\tfeatures: <' + ", ".join(d for d in args.feature_dir) + '>')
        print(f'\ttags: <{", ".join(str(g) for g in args.tags) if args.tags else "n/a"}>')

    # get all features and testable tests
    features = feature_file_parser.get_parsed_gherkin(args.feature_dir)
    features_with_tests = [f for f in features if f.get_testable_tests(args.tags)]

    # split the features into groups as evenly as possible
    tests_to_run = feature_splitter.get_grouped_tests(
        features_with_tests,
        args.tags,
        args.processes,  # max number of processes
        args.unit,  # unit to split by
    )

    logging.info('\n\t' + '\n\t'.join(str(group) for group in tests_to_run))

    # make the args available to each group
    for group in tests_to_run:
        group.extend([args, args_unknown])

    # Run the tests ---------------------------------------------------------------------------------
    # run all the processes and save the locations of the result files
    num_groups = len(tests_to_run)
    passed, failed, total = 0, 0, 0
    if args.verbose and not args.dry_run:
        print(f'\n--- RUNNING IN <{num_groups}> PROCESS{"ES" * int(num_groups > 1)}\n')

    if tests_to_run:
        # set up the multiple processes
        pool = Pool(args.processes)
        result_files = pool.map(command_generator, tests_to_run)
        logging.info(f'output files: {result_files}')

        results.write_result_aggregate(files=result_files, aggregate_out_file=args.aggregate)

        if args.dry_run:
            return 0

        # do the math to print out the results summary
        results.write_junit(args.aggregate, args.junit)
        passed, failed, total = results.get_result_values(args.aggregate)
        end = datetime.now()
        results.print_result_summary(args, args.D, start, end, passed, failed)
        root_logger.info(f'passed: {passed}')
        root_logger.info(f'failed: {failed}')

    shutil.rmtree(args.output) if args.output.endswith('.tempoutput') else None

    # exit correctly
    if args.pass_with_no_tests or not tests_to_run or total == 0:
        print('𓃥 no tests found --> this run passes by default 𓃥')
        return 0
    return 0 if passed / total >= args.pass_threshold else 1


if __name__ == '__main__':
    # run everything
    sys.exit(main())
